﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace HF_Demo
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            string filter = "Image|*.jpg;*.png;*.gif";
            ofd.Filter = filter;
            sfd.Filter = filter;

        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;
        string image_path;
        Bitmap image, processed, display;

        private void image_frame_Click(object sender, EventArgs e)
        {
            processed = image;
            MouseEventArgs me = (MouseEventArgs)e;
            Point coordinates = me.Location;
            int x = (int)((double)(coordinates.X * image.Width) / display.Width);
            int y = (int)((double)(coordinates.Y * image.Height) / display.Height);
            processed = processed.FillHoles(x, y);
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                processed = image;
                display = processed.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static Bitmap FillHoles(this Bitmap image, int start_x, int start_y)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            byte[] temp = new byte[bytes];
            byte[] result = new byte[bytes];

            //seed point
            for (int c = 0; c < 3; c++)
            {
                temp[start_x * 3 + start_y * image_data.Stride + c] = 255;
            }

            while (true)
            {
                //iterate across the image with structuring element
                for (int i = 1; i < w - 1; i++)
                {
                    for (int j = 1; j < h - 1; j++)
                    {
                        //position of the structuring element center
                        int position = i * 3 + j * image_data.Stride;
                        //set up structuring element
                        for (int k = -1; k < 2; k++)
                        {
                            for (int l = -1; l < 2; l++)
                            {
                                //position of pixel overlaped by structuring element
                                int se_pos = position + k * 3 + l * image_data.Stride;
                                if (Structuring_Element.Vals[k + 1, l + 1] == 1 && buffer[se_pos] < 255)
                                {
                                    for (int c = 0; c < 3; c++)
                                    {
                                        result[se_pos + c] = Math.Max(temp[position], result[se_pos]);
                                    }
                                    
                                }
                            }
                        }
                    }
                }
                int difference = 0;
                for (int i = 0; i < bytes; i++)
                {
                    if (result[i] != temp[i])
                    {
                        temp[i] = result[i];
                        difference++;
                    }
                }

                if (difference == 0)
                {
                    break;
                }
            }

            //combine filling pixels with input image
            for (int i = 0; i < bytes; i++)
            {
                if (result[i] < buffer[i])
                {
                    result[i] = buffer[i];
                }
            }


            Bitmap res_img = new Bitmap(w, h);
            BitmapData res_data = res_img.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);
            Marshal.Copy(result, 0, res_data.Scan0, bytes);
            res_img.UnlockBits(res_data);

            return res_img;
        }
    }

    public static class Structuring_Element
    {
        public static int[,] Vals
        {
            get
            {
                return new int[,]
                {
                    { 0, 1, 0 },
                    { 1, 1, 1 },
                    { 0, 1, 0 }
                };
            }
        }
    }
}
